;;; JACAL: Symbolic Mathematics System.        -*-scheme-*-
;;; Copyright 1989, 1990, 1991, 1992, 1993, 1995, 1997 Aubrey Jaffer.
;;; See the file "COPYING" for terms applying to this program.

;;(require 'rev3-procedures)
(require 'rev4-optional-procedures)
(require 'multiarg/and-)
(require 'with-file)
(require 'transcript)
;;; (if (and (provided? 'defmacro) (not (provided? 'macro)))
;;;     (require 'macro-by-example))
(require 'fluid-let)
(require 'common-list-functions)
(require 'sort)
;;(require 'alist)
(require 'hash-table)
(or (provided? 'inexact) (require 'logical)) ;for integer-expt
(define integer-expt (if (provided? 'inexact) expt integer-expt))

(define (copy-vector v)
  (let* ((vl (vector-length v))
	 (iv (make-vector vl)))
    (do ((i (+ -1 (vector-length iv)) (+ -1 i)))
	((negative? i))
      (vector-set! iv i (vector-ref v i)))
    iv))

(define base-language 'scheme)
;;; The rest define dummy functions for common-lisp compatability
;(define char-code-limit 256)
(define (funcall p . args) (apply p args))
(define proclaim identity)
